@@ -1,5 +1,6 @@
 __all__ = ('ForwardRefPolicy', 'TypeHintWarning', 'typechecked', 'check_return_type',
-           'check_argument_types', 'check_type', 'TypeWarning', 'TypeChecker')
+           'check_argument_types', 'check_type', 'TypeWarning', 'TypeChecker',
+           'typeguard_ignore')
 
 import collections.abc
 import gc
@@ -16,18 +17,28 @@ from types import CodeType, FunctionType
 from typing import (
     Callable, Any, Union, Dict, List, TypeVar, Tuple, Set, Sequence, get_type_hints, TextIO,
     Optional, IO, BinaryIO, Type, Generator, overload, Iterable, AsyncIterable, Iterator,
-    AsyncIterator, AbstractSet)
+    AsyncIterator, AbstractSet, TYPE_CHECKING)
 from unittest.mock import Mock
 from warnings import warn
 from weakref import WeakKeyDictionary, WeakValueDictionary
 
+# Python 3.8+
 try:
-    from typing_extensions import Literal, NoReturn
+    from typing_extensions import Literal
 except ImportError:
     try:
-        from typing import Literal, NoReturn
+        from typing import Literal
     except ImportError:
-        Literal = NoReturn = None
+        Literal = None
+
+# Python 3.5.4+ / 3.6.2+
+try:
+    from typing_extensions import NoReturn
+except ImportError:
+    try:
+        from typing import NoReturn
+    except ImportError:
+        NoReturn = None
 
 # Python 3.6+
 try:
@@ -51,11 +62,21 @@ except ImportError:
     evaluate_forwardref = ForwardRef._eval_type
 
 
+if TYPE_CHECKING:
+    _F = TypeVar("_F")
+
+    def typeguard_ignore(f: _F) -> _F:
+        """This decorator is a noop during static type-checking."""
+        return f
+else:
+    from typing import no_type_check as typeguard_ignore
+
+
 _type_hints_map = WeakKeyDictionary()  # type: Dict[FunctionType, Dict[str, Any]]
 _functions_map = WeakValueDictionary()  # type: Dict[CodeType, FunctionType]
 _missing = object()
 
-T_CallableOrType = TypeVar('T_CallableOrType', Callable, Type[Any])
+T_CallableOrType = TypeVar('T_CallableOrType', bound=Callable[..., Any])
 
 
 class ForwardRefPolicy(Enum):
@@ -84,6 +105,13 @@ class _TypeCheckMemo:
         self.typevars = {}  # type: Dict[Any, type]
 
 
+def _strip_annotation(annotation):
+    if isinstance(annotation, str):
+        return annotation.strip("'")
+    else:
+        return annotation
+
+
 class _CallMemo(_TypeCheckMemo):
     __slots__ = 'func', 'func_name', 'arguments', 'is_generator', 'type_hints'
 
@@ -116,7 +144,7 @@ class _CallMemo(_TypeCheckMemo):
 
                     typename = str(exc).split("'", 2)[1]
                     for param in signature.parameters.values():
-                        if param.annotation == typename:
+                        if _strip_annotation(param.annotation) == typename:
                             break
                     else:
                         raise
@@ -125,10 +153,11 @@ class _CallMemo(_TypeCheckMemo):
                     if forward_refs_policy is ForwardRefPolicy.GUESS:
                         if param.name in self.arguments:
                             argtype = self.arguments[param.name].__class__
-                            if param.annotation == argtype.__qualname__:
+                            stripped = _strip_annotation(param.annotation)
+                            if stripped == argtype.__qualname__:
                                 func.__annotations__[param.name] = argtype
                                 msg = ('Replaced forward declaration {!r} in {} with {!r}'
-                                       .format(param.annotation, func_name, argtype))
+                                       .format(stripped, func_name, argtype))
                                 warn(TypeHintWarning(msg))
                                 continue
 
@@ -163,7 +192,11 @@ class _CallMemo(_TypeCheckMemo):
 
 def resolve_forwardref(maybe_ref, memo: _TypeCheckMemo):
     if isinstance(maybe_ref, ForwardRef):
-        return evaluate_forwardref(maybe_ref, memo.globals, memo.locals)
+        if sys.version_info < (3, 9, 0):
+            return evaluate_forwardref(maybe_ref, memo.globals, memo.locals)
+        else:
+            return evaluate_forwardref(maybe_ref, memo.globals, memo.locals, frozenset())
+
     else:
         return maybe_ref
 
@@ -237,7 +270,7 @@ def check_callable(argname: str, value, expected_type, memo: _TypeCheckMemo) ->
     if not callable(value):
         raise TypeError('{} must be a callable'.format(argname))
 
-    if expected_type.__args__:
+    if getattr(expected_type, "__args__", None):
         try:
             signature = inspect.signature(value)
         except (TypeError, ValueError):
@@ -287,7 +320,8 @@ def check_dict(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None
                         format(argname, qualified_name(value)))
 
     if expected_type is not dict:
-        if expected_type.__args__ not in (None, expected_type.__parameters__):
+        if (hasattr(expected_type, "__args__") and
+                expected_type.__args__ not in (None, expected_type.__parameters__)):
             key_type, value_type = expected_type.__args__
             if key_type is not Any or value_type is not Any:
                 for k, v in value.items():
@@ -296,19 +330,22 @@ def check_dict(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None
 
 
 def check_typed_dict(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
-    expected_keys = frozenset(expected_type.__annotations__)
-    existing_keys = frozenset(value)
+    declared_keys = frozenset(expected_type.__annotations__)
+    if hasattr(expected_type, '__required_keys__'):
+        required_keys = expected_type.__required_keys__
+    else:  # py3.8 and lower
+        required_keys = declared_keys if expected_type.__total__ else frozenset()
 
-    extra_keys = existing_keys - expected_keys
+    existing_keys = frozenset(value)
+    extra_keys = existing_keys - declared_keys
     if extra_keys:
         keys_formatted = ', '.join('"{}"'.format(key) for key in sorted(extra_keys))
         raise TypeError('extra key(s) ({}) in {}'.format(keys_formatted, argname))
 
-    if expected_type.__total__:
-        missing_keys = expected_keys - existing_keys
-        if missing_keys:
-            keys_formatted = ', '.join('"{}"'.format(key) for key in sorted(missing_keys))
-            raise TypeError('required key(s) ({}) missing from {}'.format(keys_formatted, argname))
+    missing_keys = required_keys - existing_keys
+    if missing_keys:
+        keys_formatted = ', '.join('"{}"'.format(key) for key in sorted(missing_keys))
+        raise TypeError('required key(s) ({}) missing from {}'.format(keys_formatted, argname))
 
     for key, argtype in get_type_hints(expected_type).items():
         argvalue = value.get(key, _missing)
@@ -322,7 +359,8 @@ def check_list(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None
                         format(argname, qualified_name(value)))
 
     if expected_type is not list:
-        if expected_type.__args__ not in (None, expected_type.__parameters__):
+        if hasattr(expected_type, "__args__") and expected_type.__args__ not in \
+                (None, expected_type.__parameters__):
             value_type = expected_type.__args__[0]
             if value_type is not Any:
                 for i, v in enumerate(value):
@@ -334,7 +372,8 @@ def check_sequence(argname: str, value, expected_type, memo: _TypeCheckMemo) ->
         raise TypeError('type of {} must be a sequence; got {} instead'.
                         format(argname, qualified_name(value)))
 
-    if expected_type.__args__ not in (None, expected_type.__parameters__):
+    if hasattr(expected_type, "__args__") and expected_type.__args__ not in \
+            (None, expected_type.__parameters__):
         value_type = expected_type.__args__[0]
         if value_type is not Any:
             for i, v in enumerate(value):
@@ -347,7 +386,8 @@ def check_set(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
                         format(argname, qualified_name(value)))
 
     if expected_type is not set:
-        if expected_type.__args__ not in (None, expected_type.__parameters__):
+        if hasattr(expected_type, "__args__") and expected_type.__args__ not in \
+                (None, expected_type.__parameters__):
             value_type = expected_type.__args__[0]
             if value_type is not Any:
                 for v in value:
@@ -356,12 +396,23 @@ def check_set(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
 
 def check_tuple(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     # Specialized check for NamedTuples
-    if hasattr(expected_type, '_field_types'):
+    is_named_tuple = False
+    if sys.version_info < (3, 8, 0):
+        is_named_tuple = hasattr(expected_type, '_field_types')  # deprecated since python 3.8
+    else:
+        is_named_tuple = hasattr(expected_type, '__annotations__')
+
+    if is_named_tuple:
         if not isinstance(value, expected_type):
             raise TypeError('type of {} must be a named tuple of type {}; got {} instead'.
                             format(argname, qualified_name(expected_type), qualified_name(value)))
 
-        for name, field_type in expected_type._field_types.items():
+        if sys.version_info < (3, 8, 0):
+            field_types = expected_type._field_types
+        else:
+            field_types = expected_type.__annotations__
+
+        for name, field_type in field_types.items():
             check_type('{}.{}'.format(argname, name), getattr(value, name), field_type, memo)
 
         return
@@ -426,7 +477,9 @@ def check_class(argname: str, value, expected_type, memo: _TypeCheckMemo) -> Non
     if expected_type is Type:
         return
 
-    expected_class = expected_type.__args__[0] if expected_type.__args__ else None
+    expected_class = None
+    if hasattr(expected_type, "__args__") and expected_type.__args__:
+        expected_class = expected_type.__args__[0]
     if expected_class:
         if expected_class is Any:
             return
@@ -473,15 +526,29 @@ def check_typevar(argname: str, value, typevar: TypeVar, memo: _TypeCheckMemo,
 
 
 def check_literal(argname: str, value, expected_type, memo: _TypeCheckMemo):
-    try:
-        args = expected_type.__args__
-    except AttributeError:
-        # Instance of Literal from typing_extensions
-        args = expected_type.__values__
+    def get_args(literal):
+        try:
+            args = literal.__args__
+        except AttributeError:
+            # Instance of Literal from typing_extensions
+            args = literal.__values__
 
-    if value not in args:
+        retval = []
+        for arg in args:
+            if isinstance(arg, Literal.__class__) or getattr(arg, '__origin__', None) is Literal:
+                # The first check works on py3.6 and lower, the second one on py3.7+
+                retval.extend(get_args(arg))
+            elif isinstance(arg, (int, str, bytes, bool, type(None), Enum)):
+                retval.append(arg)
+            else:
+                raise TypeError('Illegal literal value: {}'.format(arg))
+
+        return retval
+
+    final_args = tuple(get_args(expected_type))
+    if value not in final_args:
         raise TypeError('the value of {} must be one of {}; got {} instead'.
-                        format(argname, args, value))
+                        format(argname, final_args, value))
 
 
 def check_number(argname: str, value, expected_type):
@@ -614,9 +681,10 @@ def check_type(argname: str, value, expected_type, memo: Optional[_TypeCheckMemo
 
             if expected_type is bytes:
                 # As per https://github.com/python/typing/issues/552
-                expected_type = (bytearray, bytes)
-
-            if not isinstance(value, expected_type):
+                if not isinstance(value, (bytearray, bytes, memoryview)):
+                    raise TypeError('type of {} must be bytes-like; got {} instead'
+                                    .format(argname, qualified_name(value)))
+            elif not isinstance(value, expected_type):
                 raise TypeError(
                     'type of {} must be {}; got {} instead'.
                     format(argname, qualified_name(expected_type), qualified_name(value)))
@@ -662,7 +730,7 @@ def check_return_type(retval, memo: Optional[_CallMemo] = None) -> bool:
         try:
             check_type('the return value', retval, memo.type_hints['return'], memo)
         except TypeError as exc:  # suppress unnecessarily long tracebacks
-            raise exc from None
+            raise TypeError(*exc.args) from None
 
     return True
 
@@ -697,17 +765,20 @@ def check_argument_types(memo: Optional[_CallMemo] = None) -> bool:
             try:
                 check_type(description, value, expected_type, memo)
             except TypeError as exc:  # suppress unnecessarily long tracebacks
-                raise exc from None
+                raise TypeError(*exc.args) from None
 
     return True
 
 
 class TypeCheckedGenerator:
     def __init__(self, wrapped: Generator, memo: _CallMemo):
-        rtype_args = memo.type_hints['return'].__args__
+        rtype_args = []
+        if hasattr(memo.type_hints['return'], "__args__"):
+            rtype_args = memo.type_hints['return'].__args__
+
         self.__wrapped = wrapped
         self.__memo = memo
-        self.__yield_type = rtype_args[0]
+        self.__yield_type = rtype_args[0] if rtype_args else Any
         self.__send_type = rtype_args[1] if len(rtype_args) > 1 else Any
         self.__return_type = rtype_args[2] if len(rtype_args) > 2 else Any
         self.__initialized = False
@@ -797,8 +868,9 @@ def typechecked(func=None, *, always=False, _localns: Optional[Dict[str, Any]] =
     If the ``__debug__`` global variable is set to ``False``, no wrapping and therefore no type
     checking is done, unless ``always`` is ``True``.
 
-    This can also be used as a class decorator. This will wrap all type annotated methods in the
-    class with this decorator.
+    This can also be used as a class decorator. This will wrap all type annotated methods,
+    including ``@classmethod``, ``@staticmethod``,  and ``@property`` decorated methods,
+    in the class with the ``@typechecked`` decorator.
 
     :param func: the function or class to enable type checking for
     :param always: ``True`` to enable type checks even in optimized mode
@@ -820,6 +892,17 @@ def typechecked(func=None, *, always=False, _localns: Optional[Dict[str, Any]] =
                 if getattr(attr.__func__, '__annotations__', None):
                     wrapped = typechecked(attr.__func__, always=always, _localns=func.__dict__)
                     setattr(func, key, type(attr)(wrapped))
+            elif isinstance(attr, property):
+                kwargs = dict(doc=attr.__doc__)
+                for name in ("fset", "fget", "fdel"):
+                    property_func = getattr(attr, name)
+                    if property_func is None:
+                        continue
+                    kwargs[name] = typechecked(
+                        property_func, always=always, _localns=func.__dict__
+                    )
+
+                setattr(func, key, property(**kwargs))
 
         return func
 
@@ -838,7 +921,10 @@ def typechecked(func=None, *, always=False, _localns: Optional[Dict[str, Any]] =
         memo = _CallMemo(python_func, _localns, args=args, kwargs=kwargs)
         check_argument_types(memo)
         retval = func(*args, **kwargs)
-        check_return_type(retval, memo)
+        try:
+            check_return_type(retval, memo)
+        except TypeError as exc:
+            raise TypeError(*exc.args) from None
 
         # If a generator is returned, wrap it if its yield/send/return types can be checked
         if inspect.isgenerator(retval) or isasyncgen(retval):
